import tkinter as tk
import customtkinter as ctk
from hydroecolstm.model.lstm_linears import Lstm_Linears
from hydroecolstm.model.ea_lstm import Ea_Lstm_Linears
from hydroecolstm.train.trainer import Trainer
from CTkToolTip import CTkToolTip
import torch

class TrainTestFrame(ctk.CTkScrollableFrame):
    def __init__(self, container=None, config=None, globalData=None):
        super().__init__(container)
        
        # setup the grid layout manager
        self.config = config
        self.globalData = globalData
        self.__create_widgets() 
        
    # create widgets for sidebar frame 
    def __create_widgets(self): 
        
        # create tabs #
        self.tabview = ctk.CTkTabview(master=self, width = 750, border_width=1.5,
                                      fg_color = "transparent")
        self.tabview.pack(fill='both',expand=1)
        self.tabview.add("1. Initial state dicts")
        self.tabview.tab("1. Initial state dicts").grid_columnconfigure((0), weight=1)
        self.tabview.add("2. Trainer")
        self.tabview.tab("2. Trainer").grid_columnconfigure((0,1), weight=1)
        
        # ---------------------------------------------Initial state dicts      
        # Load model state dict
        self.load_model_label = ctk.CTkLabel(self.tabview.tab("1. Initial state dicts"), 
                                      text="1. Load model state dicts")
        self.load_model_label.grid(row=0, column=0, sticky = "w")  
        self.load_model = ctk.CTkButton(self.tabview.tab("1. Initial state dicts"), anchor='w', 
                                         command=self.load_state_dict,
                                         text="Load model")
        self.load_model.grid(row=1, column=0, sticky = "w")
        CTkToolTip(self.load_model, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="Optional input - load initial model state dicts. " +
                   "This could be, e.g., the calibrated model at a regional scale " +
                   "and used in here for parameter fine tuning.")
        
        # ---------------------------------------------Content of load data tab
        # Number of epochs
        self.nepoch_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                         text="1. Number of epochs")
        self.nepoch_label.grid(row=0, column=0, sticky = "w")
        
        self.nepoch = ctk.CTkEntry(master=self.tabview.tab("2. Trainer"),
                             placeholder_text="5")
               
        self.nepoch.grid(row=1, column=0, sticky = "w")
        self.nepoch.bind('<KeyRelease>', self.get_nepoch)
        CTkToolTip(self.nepoch, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message='Number of training epochs (positive integer number)') 
        
        # Learning rate
        self.learning_rate_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                                text="2. Learning rate")
        self.learning_rate_label.grid(row=2, column=0, sticky = "w")
        self.learning_rate= ctk.CTkEntry(master=self.tabview.tab("2. Trainer"),
                             placeholder_text="0.01")
        self.learning_rate.grid(row=3, column=0, sticky = "w")
        self.learning_rate.bind('<KeyRelease>', self.get_learning_rate)
        CTkToolTip(self.learning_rate, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message='Learning rate for gradient descent (positive real number)') 
        
        # Warm up length
        self.warmup_length_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                         text="3. Warm-up length")
        self.warmup_length_label.grid(row=4, column=0, sticky = "w")
        self.warmup_length= ctk.CTkEntry(master=self.tabview.tab("2. Trainer"),
                             placeholder_text="30") 
        self.warmup_length.grid(row=5, column=0, sticky = "w")
        self.warmup_length.bind('<KeyRelease>', self.get_warmup_length)
        CTkToolTip(self.warmup_length, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message='The number of timesteps used for warm-up. \n' +
                   'For example, the first warmup_length outputs will be skipped \n'+
                   'when calculating the loss function. This value MUST \n'+
                   'be smaller than sequence length') 
        
        # Sequence length
        self.sequence_length_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                         text="4. Sequence length")
        self.sequence_length_label.grid(row=6, column=0, sticky = "w")
        self.sequence_length= ctk.CTkEntry(master=self.tabview.tab("2. Trainer"),
                             placeholder_text="720") 
        self.sequence_length.grid(row=7, column=0, sticky = "w")
        self.sequence_length.bind('<KeyRelease>', self.get_sequence_length)
        CTkToolTip(self.sequence_length, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="The number of timesteps in each sample dataset. \n" + 
                   "One epoch loops over multiple baches. \n" +
                   "One batch consists of 'batch_size' sample datasets. \n" +
                   "One sample dataset consist of pairs of Input and Tartget \n" +
                   "    of 'sequence_length' timesteps (chronological order)")

        # Batch size
        self.batch_size_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                         text="5. Batch size")
        self.batch_size_label.grid(row=8, column=0, sticky = "w")
        self.batch_size= ctk.CTkEntry(master=self.tabview.tab("2. Trainer"),
                             placeholder_text="3") 
        self.batch_size.grid(row=9, column=0, sticky = "w")
        self.batch_size.bind('<KeyRelease>', self.get_batch_size)
        CTkToolTip(self.batch_size, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="Please see sequence length for help") 
        
        # Patience
        self.patience_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                         text="6. Patience length")
        self.patience_label.grid(row=0, column=2, sticky = "w")
        self.patience= ctk.CTkEntry(master=self.tabview.tab("2. Trainer"),
                             placeholder_text="20") 
        self.patience.grid(row=1, column=2, sticky = "w")
        self.patience.bind('<KeyRelease>', self.get_patience_length)
        CTkToolTip(self.patience, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="The number of epochs to wait (before stopping) to \n" + 
                   "see if there is no improvement in ths validation loss, \n" + 
                   "which is used for early stopping. More information \n" + 
                   "please see https://github.com/Bjarten/early-stopping-pytorch. \n" +
                   "Patience length should be much smaller than the number of epochs")

        # Optimization method
        self.optim_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                        text="7. Optimization method")
        self.optim_label.grid(row=2, column=2, sticky = "w")    
        self.optim = ctk.CTkOptionMenu(self.tabview.tab("2. Trainer"),
                                                   values=['Adam'],
                                                   command=self.get_optim_method) 
        self.optim.grid(row=3, column=2, sticky = "w")
        CTkToolTip(self.optim, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="In this version only Adam method is available")
        
        # Loss function
        self.loss_function_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                               text="8. Loss function")
        self.loss_function_label.grid(row=4, column=2, sticky = "w")    
        self.loss = ctk.CTkOptionMenu(self.tabview.tab("2. Trainer"),
                                                   values=['Root Mean Square Error',
                                                           'Mean Absolute Error',
                                                           'Mean Squared Error'],
                                                   command=self.loss_function) 
        self.loss.grid(row=5, column=2, sticky = "w")
        CTkToolTip(self.loss, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="Please select the loss function for calculating"+
                   " traning and validatation losses, which are used for" +
                   " updating model paramters and early stopping")
        
        # Save model
        self.out_dir_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                      text="9. Output directory")
        self.out_dir_label.grid(row=6, column=2, sticky = "w")     
        self.out_dir = ctk.CTkButton(self.tabview.tab("2. Trainer"), anchor='w', 
                                         command=self.out_dir_event,
                                         text="Select directory")
        self.out_dir.grid(row=7, column=2, sticky = "w")
        CTkToolTip(self.out_dir, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="Please select the directory for saving model outputs"+
                   " If not selected, the default directory (pathlib.Path.cwd()) is used")
        
        
        # Run model
        self.run_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                      text="10. Run model")
        self.run_label.grid(row=8, column=2, sticky = "w")     
        self.run = ctk.CTkButton(self.tabview.tab("2. Trainer"), anchor='w', 
                                         command=self.run_train_test,
                                         text="Run")
        self.run.grid(row=9, column=2, sticky = "w")
        CTkToolTip(self.run, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="Click here to run the model. While model is running"+
                   " in the background, the GUI will be frozen.")
                
        # Progressbar
        self.progressbar = ctk.CTkProgressBar(master=self.tabview.tab("2. Trainer"))
        self.progressbar.grid(row=10, column=2,  sticky = "w", pady = (10,10))
        CTkToolTip(self.progressbar, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="This is the training progress bar." +
                   " It will turn to full orange when the training was completed")
        
        self.progressbar.configure(mode="determinate", progress_color="orange")
        self.progressbar.set(0)
        
    # Get number of epochs
    def get_nepoch(self, dummy):
        try:
            self.config["n_epochs"] = int(self.nepoch.get().strip())
            print(f"Number of epochs = {self.config['n_epochs']}")
        except:
            tk.messagebox.showinfo(title="Error", message="Input should be integer")
        
    # Get learning_rate
    def get_learning_rate(self, dummy):
        try:
            self.config["learning_rate"]  = float(self.learning_rate.get().strip())
            print(f"Learning rate = {self.config['learning_rate']}")
        except:
            tk.messagebox.showinfo(title="Error", message="Input should be numeric")

    # Get warm up length
    def get_warmup_length(self, dummy):
        try:
            self.config["warmup_length"] =  int(self.warmup_length.get().strip()) 
            print(f"Warm-up length = {self.config['warmup_length']}")  
        except:
            tk.messagebox.showinfo(title="Error", message="Input should be integer")
     
    # Get learning_rate
    def get_sequence_length(self, dummy):
        try:
            self.config["sequence_length"]  = int(self.sequence_length.get().strip())
            print(f"Sequence length = {self.config['sequence_length']}")
        except:
            tk.messagebox.showinfo(title="Error", message="Input should be integer")

    # Get learning_rate
    def get_batch_size(self, dummy):
        try:
            self.config["batch_size"]  = int(self.batch_size.get().strip())
            print(f"Batch size = {self.config['batch_size']}")
        except:
            tk.messagebox.showinfo(title="Error", message="Input should be integer")

    # Get learning_rate
    def get_patience_length(self, dummy):
        try:
            self.config["patience"]  = int(self.patience.get().strip())
            print(f"Patience = {self.config['patience']}")
        except:
            tk.messagebox.showinfo(title="Error", message="Input should be integer")
        
    # Get number of lstm layers
    def loss_function(self, method: str):
        loss_fn = {'Root Mean Square Error': "RMSE",
                   'Mean Absolute Error': "MAE",
                   'Mean Squared Error': "MSE"} 
                              
        self.config["loss_function"] = loss_fn[method]
        print(self.config["loss_function"])

    # Get number of lstm layers
    def get_optim_method(self, method: str):
        self.config["optim_method"] = method
        print(self.config["optim_method"])
        
    def out_dir_event(self):
        output_directory = tk.filedialog.askdirectory()
        self.config["output_directory"] = [output_directory]
        print("Output dir = ", self.config["output_directory"])
        
        # update text of the selected directory
        if len(output_directory) > 0:
            self.out_dir.configure(text = "..." + output_directory[-30:])
        else:
            self.out_dir.configure(text = "Select directory")
    
    def load_state_dict(self):
        # Select model state dict file
        state_dict_file = ctk.filedialog.askopenfilename(title="Select model state dict file", 
                                              filetypes=(('pt files', '*.pt'),
                                                         ('All files', '*.*')))
        # update text load model
        if state_dict_file[-3:] == ".pt":
            self.load_model.configure(text = "..." + state_dict_file[-30:])
            self.globalData['init_state_dicts'] = True
            self.globalData['init_state_dicts_file'] = state_dict_file
        else:
            self.load_model.configure(text = "Load model")
            self.globalData['init_state_dicts'] = False

     
    def run_train_test(self):
        # Set progress to zero
        self.progressbar.set(0)

        self.run.configure(fg_color='gray')
        self.run.configure(state="disabled")
        
        tk.messagebox.showinfo(title="Message box", 
                               message="Trainning will start after closing this box")
        
        try:
            if self.config["model_class"] == "LSTM":
                self.globalData["model"] = Lstm_Linears(config=self.config)
            else:
                self.globalData["model"] = Ea_Lstm_Linears(config=self.config)
                
            print("done create model")
            
            # Initialize weights, biases
            if self.globalData['init_state_dicts']:
                self.globalData["model"].load_state_dict(
                    torch.load(self.globalData['init_state_dicts_file']))
                
            # Train the model
            self.globalData["trainer"] = Trainer(config=self.config, 
                                               model=self.globalData["model"])
            
            
            print("done initialize trainer")
            self.globalData["model"] = self.globalData["trainer"].train(
                self.globalData["x_train_scale"],
                self.globalData["y_train_scale"],
                self.globalData["x_valid_scale"],
                self.globalData["y_valid_scale"])
            print("done train model")
            # Run forward test the model
            y_train_simulated_scale = self.globalData["model"].evaluate(
                self.globalData["x_train_scale"])
            y_valid_simulated_scale = self.globalData["model"].evaluate(
                self.globalData["x_valid_scale"])
            y_test_simulated_scale = self.globalData["model"].evaluate(
                self.globalData["x_test_scale"])

            # Inverse scale/transform back simulated result to real scale
            self.globalData["y_train_simulated"] =\
                self.globalData["y_scaler"].inverse(y_train_simulated_scale)
            self.globalData["y_valid_simulated"] =\
                self.globalData["y_scaler"].inverse(y_valid_simulated_scale)
            self.globalData["y_test_simulated"] =\
                self.globalData["y_scaler"].inverse(y_test_simulated_scale)
                
            tk.messagebox.showinfo(title="Message box",
                                   message="Finished training/testing")
        except:
            tk.messagebox.showinfo(title="Error",
                                   message="Cannot train the model")

        self.progressbar.set(1.0)        
        self.run.configure(state="normal")
        self.run.configure(fg_color=['#3a7ebf', '#1f538d'])

    
        
        
        


#!/usr/bin/env python

from hydroecolstm.data.read_data import read_train_valid_test_data
from hydroecolstm.data.scaler import Scaler, get_scaler_name
from hydroecolstm.model.lstm_linears import Lstm_Linears
from hydroecolstm.model.ea_lstm import Ea_Lstm_Linears
from hydroecolstm.train.trainer import Trainer

# Function to train and test the model 
def run_train(config):

    # Read and split data
    data = read_train_valid_test_data(config)
    
    # Scale/transformer name for static, dynamic, and target features
    x_scaler_name, y_scaler_name = get_scaler_name(config)
    
    # Scaler/transformer
    x_scaler, y_scaler = Scaler(), Scaler()
    x_scaler.fit(x=data["x_train"], method=x_scaler_name)
    y_scaler.fit(x=data["y_train"], method=y_scaler_name)
    
    # Scale/transform data input data
    x_train_scale = x_scaler.transform(x=data["x_train"])
    x_valid_scale = x_scaler.transform(x=data["x_valid"])
    x_test_scale = x_scaler.transform(x=data["x_test"])

    # Scale/transform data target train data    
    y_train_scale = y_scaler.transform(x=data["y_train"])
    y_valid_scale = y_scaler.transform(x=data["y_valid"])
    
    # Create the model
    if config["model_class"] == "LSTM":
        model = Lstm_Linears(config)
    else:
        model = Ea_Lstm_Linears(config)
        
    # Train with train dataset
    trainer = Trainer(config, model)
    model = trainer.train(x_train_scale, y_train_scale, x_valid_scale, y_valid_scale)
    
    # Run with input with Dict[str:torch.Tensor] and torch.no_grad()
    y_train_simulated_scale = model.evaluate(x_train_scale)
    y_valid_simulated_scale = model.evaluate(x_valid_scale)
    y_test_simulated_scale = model.evaluate(x_test_scale)
    
    # Inverse scale/transform back simulated result to real scale
    data["y_train_simulated"] = y_scaler.inverse(y_train_simulated_scale)
    data["y_valid_simulated"] = y_scaler.inverse(y_valid_simulated_scale)
    data["y_test_simulated"] = y_scaler.inverse(y_test_simulated_scale)
    data["trainer"] = trainer
    data["y_scaler"] = y_scaler
    data["x_scaler"] = x_scaler
    
    return model, x_scaler, y_scaler, data



import customtkinter as ctk
import tkinter as tk
import pandas as pd
from pandastable import Table
import tkcalendar as tkc
from CTkListbox import CTkListbox
from CTkToolTip import CTkToolTip
from hydroecolstm.data.scaler import Scaler, get_scaler_name
from hydroecolstm.data.read_data import read_train_valid_test_data
   
class DataFrame(ctk.CTkScrollableFrame):
    def __init__(self, container=None, config=None, globalData=None):
        super().__init__(container)
        
        # setup the grid layout manager 
        self.config = config
        self.globalData = globalData
        self.__create_widgets()
        
    # create widgets for sidebar frame
    def __create_widgets(self): 
        
        # create tabs      
        self.tabview = ctk.CTkTabview(master=self, width = 750, border_width=1.5,
                                      fg_color = "transparent")
        self.tabview.pack(fill='both',expand=1)
        self.tabview.add("1. Load data")
        self.tabview.tab("1. Load data").grid_columnconfigure((1,1), weight=1)
        self.tabview.add("2. Filter data")
        self.tabview.tab("2. Filter data").grid_columnconfigure((1,1), weight=1)
        self.tabview.add("3. Transform data")
        self.tabview.tab("3. Transform data").grid_columnconfigure((1,1), weight=1)
        
        # ---------------------------------------------content of load data tab
        # ---------------------------------------------------------Dynamic data
        self.dynamic_label = ctk.CTkLabel(self.tabview.tab("1. Load data"), 
                                         text="1. Dynamic/time series data")
        self.dynamic_label.grid(row=0, column=0, padx = 10, pady=(5,5), sticky="w")
        
        # Select dynamic data label
        self.select_dynamic_file = ctk.CTkButton(self.tabview.tab("1. Load data"), 
                                               anchor='w', 
                                               text="Select dynamic data file", 
                                               command=self.get_dynamic_file)
        self.select_dynamic_file.grid(row=1, column=0, padx = 10, pady=(5,5), sticky="w")        
        CTkToolTip(self.select_dynamic_file, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='e',  wraplength=500, 
                   message='Click here to select dynamic (timeseries) data file. ' + 
                   'The data file must be in .csv format with header. ' +
                   'This data file must contains at least two colums, ' +
                   'one with the name <object_id> and the other column with the ' +
                   'name <time> in format yyyy-mm-dd hh:mm.The object_id column will '+ 
                   'link object in dynamic/timeseries data file with static data file. ' +
                   'Timeseries of BOTH input and target features must be in this file. ' +
                   'Template of this file can be seen in the package documentation')
        
        # display selected data file
        self.selected_dynamic_filename = ctk.CTkLabel(self.tabview.tab("1. Load data"),
                                                      text="No file was selected")
        self.selected_dynamic_filename.grid(row=2, column=0, padx = 10, pady=(5,5), sticky="w")
        
        #  display dynamic data label
        self.show_data_statistics = ctk.CTkButton(self.tabview.tab("1. Load data"), 
                                              anchor='e', 
                                              text="Display/visualize dynamic data", 
                                              command=self.display_dynamic_data)
        self.show_data_statistics.grid(row=3, column=0, padx = 10, 
                                   pady=(5,5), sticky="w")        
        
        
        # check box dynamic data
        self.check_dynamic_data = ctk.IntVar(value=0)         
        self.checkbox_dynamic_data = ctk.CTkCheckBox(self.tabview.tab("1. Load data"), 
                                               text="Check/uncheck to use/discard data",
                                               command=self.load_dynamic_data, 
                                               variable=self.check_dynamic_data)
        self.checkbox_dynamic_data.grid(row=4, column=0, padx = 10, pady=(10,10), sticky="w")
        CTkToolTip(self.checkbox_dynamic_data, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'If you check this box, data from the selected \n' +
                       'dynamic data file will be used, uncheck means data will \n' +
                       'not be used')

        # ---------------------------------------------------------Static data       
        self.static_label = ctk.CTkLabel(self.tabview.tab("1. Load data"), 
                                        text="2. Static attribute data")
        self.static_label.grid(row=0, column=2, padx = 10, pady=(5,5), sticky="w")
        
        # select static data file
        self.select_static_file = ctk.CTkButton(self.tabview.tab("1. Load data"), 
                                              anchor='e', 
                                              text="Select static data file", 
                                              command=self.get_static_file)

        self.select_static_file.grid(row=1, column=2, padx = 10, 
                                   pady=(5,5), sticky="w")
        CTkToolTip(self.select_static_file, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message='Click here to select static (catchment attributes) data file. ' + 
                   'the data file must be in .csv format with header. ' + 
                   'This data file must contains a column name <object_id> ' +
                   'to connect static with dynamic data. ' +
                   'Template of this file can be seen in the package documentation')
        
        # display selected data file
        self.selected_static_filename = ctk.CTkLabel(self.tabview.tab("1. Load data"),
                                                      text="No file was selected")
        self.selected_static_filename.grid(row=2, column=2, padx = 10, pady=(5,5), sticky="w")
        
        # display visualize static data
        self.show_data = ctk.CTkButton(self.tabview.tab("1. Load data"), 
                                              anchor='e', 
                                              text="Display/visualize static data", 
                                              command=self.display_static_data)
        self.show_data.grid(row=3, column=2, padx = 10, 
                                   pady=(5,5), sticky="w")
        CTkToolTip(self.show_data, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to have a look at the data (using pandastable)' +
                   'here you can also visualize the data with differnt type of plots')
        
        # check to load/unload
        self.check_static_data = ctk.IntVar(value=0)        
        self.checkbox_static_data = ctk.CTkCheckBox(self.tabview.tab("1. Load data"), 
                                               text="Check/uncheck to use/discard data",
                                               command=self.load_static_data, 
                                               variable=self.check_static_data)
        
        self.checkbox_static_data.grid(row=4, column=2, padx = 10, pady=(5,5), sticky="w")
        CTkToolTip(self.checkbox_static_data, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'If you check this box, data from the selected \n' +
                       'static data file will be used, uncheck means data will \n' +
                       'not be used')
        #--------------------------------------------Content of filter data tab
        # Select input features
        self.select_input_feature_label = ctk.CTkLabel(self.tabview.tab("2. Filter data"),
                                                       text="1. Select input features")
        self.select_input_feature_label.grid(row=0, column=0, padx = 10, pady=(5,5), sticky="w")
        self.select_input_feature = CTkListbox(master=self.tabview.tab("2. Filter data"), 
                                               multiple_selection=True, border_width=1.5,
                                               text_color="black")
        self.select_input_feature.grid(row=1, column=0, padx = 10, pady=(5,5), sticky="w")    
        CTkToolTip(self.select_input_feature_label, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Select input features for the neural network. \n' +
                       'You can select as many input features as you want')
        
        # Select target features
        self.select_target_feature_label = ctk.CTkLabel(self.tabview.tab("2. Filter data"),
                                                       text="3. Select target features")
        self.select_target_feature_label.grid(row=0, column=2, padx = 10, pady=(5,5), sticky="w")
        CTkToolTip(self.select_target_feature_label, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Select target features (model outputs). \n' +
                       'You can select as many target features as you want')
        self.select_target_feature = CTkListbox(master=self.tabview.tab("2. Filter data"), 
                                                multiple_selection=True, border_width=1.5,
                                               text_color="black")
        self.select_target_feature.grid(row=1, column=2, padx = 10, pady=(5,5), sticky="w") 
        
        # Select object id
        self.object_id_label = ctk.CTkLabel(self.tabview.tab("2. Filter data"),
                                                   text="2. Select object id")
        self.object_id_label.grid(row=2, column=0, padx = 10, pady=(5,5), sticky="w")
        CTkToolTip(self.object_id_label, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Select object id (basin id) that you want to put into the model. \n' +
                       'You can select as many object ids as you want')
        self.object_id = CTkListbox(master=self.tabview.tab("2. Filter data"), 
                                           multiple_selection=True, border_width=1.5,
                                           text_color="black")
        self.object_id.grid(row=3, column=0, rowspan=7, padx = 10, pady=(5,5), sticky="w") 
        
        # Trainning period
        self.select_date_train= ctk.CTkLabel(self.tabview.tab("2. Filter data"),
                                             text="4. Training period (yyyy-mm-dd)")
        self.select_date_train.grid(row=2, column=2, padx = 10, pady=(5,5), sticky="w")    
        self.start_train = tkc.DateEntry(self.tabview.tab("2. Filter data"), 
                                         date_pattern= 'yyyy-mm-dd', width = 25,
                                         year=1800, month=1, day=1, font=ctk.CTkFont(size=14))
        self.start_train.grid(row= 3,column=2, padx=30, pady=10, sticky='e')
        self.end_train = tkc.DateEntry(self.tabview.tab("2. Filter data"), 
                                       date_pattern= 'yyyy-mm-dd', width = 25,
                                       year=2010, month=1, day=1, font=ctk.CTkFont(size=14))
        self.end_train.grid(row= 4,column=2, padx=30, pady=10, sticky='e')   
        CTkToolTip(self.select_date_train, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w', 
                   message='Select starting date (upper calender box) and \n' + 
                   'ending date (lower calendar box) of the training period. \n'+
                   'Training data are used to fit model parameters.')

        # start validation calander  
        self.select_date_valid= ctk.CTkLabel(self.tabview.tab("2. Filter data"),
                                            text="5. Validation period (yyyy-mm-dd)")
        self.select_date_valid.grid(row=5, column=2, padx = 10, pady=(5,5), sticky="w")    
        self.start_valid = tkc.DateEntry(self.tabview.tab("2. Filter data"), 
                                         date_pattern= 'yyyy-mm-dd', width = 25,
                                         year=1800, month=1, day=1, font=ctk.CTkFont(size=14))
        CTkToolTip(self.select_date_valid, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w', 
                   message='Select starting date (upper calender box) and \n' + 
                   'ending date (lower calendar box) of the validtion period. \n'+
                   'Validation data ared used for early stopping.')
        self.start_valid.grid(row= 6,column=2, padx=30, pady=10, sticky='e')
        self.end_valid = tkc.DateEntry(self.tabview.tab("2. Filter data"), 
                                       date_pattern= 'yyyy-mm-dd', width = 25,
                                       year=2018, month=1, day=1, font=ctk.CTkFont(size=14))
        self.end_valid.grid(row= 7,column=2, padx=30, pady=10, sticky='e')   
        
        # start testing calander  
        self.select_date_test= ctk.CTkLabel(self.tabview.tab("2. Filter data"),
                                            text="6. Testing period (yyyy-mm-dd)")
        self.select_date_test.grid(row=8, column=2, padx = 10, pady=(5,5), sticky="w")    
        self.start_test = tkc.DateEntry(self.tabview.tab("2. Filter data"), 
                                         date_pattern= 'yyyy-mm-dd', width = 25,
                                         year=1800, month=1, day=1, font=ctk.CTkFont(size=14))
        CTkToolTip(self.select_date_test, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w', 
                   message='Select starting date (upper calender box) and \n' + 
                   'ending date (lower calendar box) of the testing period. \n' +
                   'Testing data are used for final evaluation of the model.')
        self.start_test.grid(row= 9,column=2, padx=30, pady=10, sticky='e')
        self.end_test = tkc.DateEntry(self.tabview.tab("2. Filter data"), 
                                       date_pattern= 'yyyy-mm-dd', width = 25,
                                       year=2018, month=1, day=1, font=ctk.CTkFont(size=14))
        self.end_test.grid(row= 10,column=2, padx=30, pady=10, sticky='e')   
               
        #-----------------------------------------------------3. Transform data
        self.transform_dd_label = ctk.CTkLabel(self.tabview.tab("3. Transform data"),
                                               text="1. Transformer for dynamic features")
        self.transform_dd_label.grid(row=0, column=0, pady = 3, padx = 10, sticky = "w")
        
        self.transform_dd_option = ctk.CTkOptionMenu(self.tabview.tab("3. Transform data"),
                                                   values=['MinMaxScaler', 'Z-score', 'None'],
                                                   command=self.transform_dynamic_data_option)
        self.transform_dd_option.grid(row=1, column=0, pady = 3, padx = 10, sticky = "w")
        CTkToolTip(self.transform_dd_option, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to select method for transforming dynamic data') 


        self.transform_tar_label = ctk.CTkLabel(self.tabview.tab("3. Transform data"),
                                               text="2. Transformer for target features")
        self.transform_tar_label.grid(row=4, column=0, pady = 3, padx = 10, sticky = "w")
        
        self.transform_tar_option = ctk.CTkOptionMenu(self.tabview.tab("3. Transform data"),
                                                      values=['MinMaxScaler', 'Z-score', 'None'],
                                                      command=self.transform_target_data_option) 
        self.transform_tar_option.grid(row=5, column=0, pady = (3,35), padx = 10, sticky = "w")
        CTkToolTip(self.transform_tar_option, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to select method for transforming target features') 
        
        self.transform_ss_label = ctk.CTkLabel(self.tabview.tab("3. Transform data"),
                                       text="3. Transformer for static features")
        self.transform_ss_label.grid(row=2, column=0, pady = 3, padx = 10, sticky = "w")

        self.transform_ss_option = ctk.CTkOptionMenu(self.tabview.tab("3. Transform data"),
                                                 values=['MinMaxScaler', 'Z-score', 'None'],
                                                 command=self.transform_static_data_option) 
        self.transform_ss_option.grid(row=3, column=0, pady = 3, padx = 10, sticky = "w")
        CTkToolTip(self.transform_ss_option, delay=0.1, bg_color = 'orange',
           text_color = 'black', anchor = 'w', 
           message= 'Click here to select method for transforming static data') 


        self.execute_label = ctk.CTkLabel(self.tabview.tab("3. Transform data"),
                                               text="4. Data split and transform")
        self.execute_label.grid(row=0, column=2, pady = 3, padx = 10, sticky = "w")
        
        self.execute = ctk.CTkButton(self.tabview.tab("3. Transform data"),
                                                   text="Execute",
                                                   command=self.data_split_transform) 
        self.execute.grid(row=1, column=2, pady = 3, padx = 10, sticky = "w")
        CTkToolTip(self.execute, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to split data (train, valid, test) and transform') 
        
        self.show_result_label = ctk.CTkLabel(self.tabview.tab("3. Transform data"),
                                               text="5. Display/Visualize transform data")
        self.show_result_label.grid(row=2, column=2, pady = 3, padx = 10, sticky = "w")
        
        self.show_result = ctk.CTkButton(self.tabview.tab("3. Transform data"),
                                                   text="Display/Visualize",
                                                   command=self.display_orig_trans_data) 
        self.show_result.grid(row=3, column=2, pady = 3, padx = 10, sticky = "w")
        CTkToolTip(self.show_result, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to view the original and transform data of the first object id') 

    #-----------------------------------------------------functions for widgets
    # get dynamic data file name
    def get_dynamic_file(self):
        # No file was selected when just click this button without selecting file
        self.globalData['dynamic_data_file'] = []
        self.globalData['dynamic_data'] = []
        self.globalData['dynamic_data_header'] = []
        self.globalData['object_id'] = []
        
        try:
            # get file name
            file_name = ctk.filedialog.askopenfilename(title="Select dynamic/time series data file", 
                                                       filetypes=(('csv files', '*.csv'),
                                                                  ('All files', '*.*')))
            self.globalData['dynamic_data_file'] = [file_name]
            
            # get data
            self.globalData['dynamic_data'] = pd.read_csv(file_name,
                                                          delimiter=",", header=0)
            
            # get data
            self.globalData['dynamic_data_header'] = list(self.globalData['dynamic_data'].columns)
            self.globalData['object_id'] = list(pd.unique(self.globalData['dynamic_data']["object_id"]))

            # update label
            self.selected_dynamic_filename.configure(text= '...' + file_name[-30:])
            
        except:
            None
            # update label
            self.selected_dynamic_filename.configure(text= 'No file was selected')
        
        # Show message box
        #tk.messagebox.showinfo(title="Done reading data", message="Close this box")   
      
    # show dynamic data in pandas data table
    def display_dynamic_data(self):
        try:
            self.table = Table(tk.Toplevel(self), dataframe=self.globalData['dynamic_data'], 
                               showtoolbar=True, showstatusbar=True)
            self.table.show()
        except:
            None

    # get static data file
    def get_static_file(self):
        
        self.globalData['static_data_file'] = []
        self.globalData['static_data'] = []
        self.globalData['static_data_header'] = []
        
        try:
            file_name = ctk.filedialog.askopenfilename(title="Select static data file", 
                                                  filetypes=(('csv files', '*.csv'),
                                                             ('All files', '*.*')))
            self.globalData['static_data_file'] = [file_name]
            self.globalData['static_data'] = pd.read_csv(file_name,
                                                          delimiter=",", header=0)
            self.globalData['static_data_header'] = list(self.globalData['static_data'].columns)
            
            # update label
            self.selected_static_filename.configure(text= '...' + file_name[-30:])
        except:
            None
            
            # update label
            self.selected_static_filename.configure(text= 'No file was selected')
            
    # show static data in pandas data table
    def display_static_data(self):
        try:
            self.table = Table(tk.Toplevel(self), dataframe=self.globalData['static_data'], 
                               showtoolbar=True, showstatusbar=True)
            self.table.show()  
        except:
            None    

    # Get the data
    def load_dynamic_data(self): 

        try:
            self.select_input_feature.delete(index=0, last='END')
            self.select_target_feature.delete(index=0, last='END')
            self.object_id.delete(index=0, last='END')
        except:
            None
            
        # Get the static data file name
        if self.check_dynamic_data.get() == 0:
            try: 
                del self.config['dynamic_data_file']
            except:
                None
            
        else:
            self.config['dynamic_data_file'] = self.globalData['dynamic_data_file']
            
        try:
            if (("dynamic_data_file" in self.config) & 
                (self.check_dynamic_data.get() == 1)):
                items = self.globalData['dynamic_data_header'].copy()
                object_id = self.globalData['object_id'].copy()
                 
                if (("static_data_file" in self.config) & 
                    (self.check_static_data.get() == 1)):
                    items.extend(self.globalData['static_data_header'])

                # remove columns with object_id and time
                items = [i for i in items if i != "object_id"]
                items = [i for i in items if i != "time"]
                
                for i in items:
                    self.select_input_feature.insert('END', option=i)
                    self.select_target_feature.insert('END', option=i)
                    
                for i in object_id:
                    self.object_id.insert('END', option=i)
            else:
                if (("static_data_file" in self.config) & 
                    (self.check_static_data.get() == 1)):
                    items = self.globalData['static_data_header'].copy()
                     
                    # remove columns with object_id and time
                    items = [i for i in items if i != "object_id"]
                    items = [i for i in items if i != "time"]
                    
                    for i in items:
                        self.select_input_feature.insert('END', option=i)
                        self.select_target_feature.insert('END', option=i)
        except:
            None

           
    def load_static_data(self): 
        
        # First remove content of the select input and target features box
        try:
            self.select_input_feature.delete(index=0, last='END')
            self.select_target_feature.delete(index=0, last='END')

        except:
            None
        
        # Get the static data file name
        if self.check_static_data.get() == 0:
            try: 
                del self.config['static_data_file']
            except:
                None
        else:
            self.config['static_data_file'] = self.globalData['static_data_file']
            
        # Get content for select input and target feature box
        try:
            if (("static_data_file" in self.config) & 
                (self.check_static_data.get() == 1)):
                items = self.globalData['static_data_header'].copy()

                if (("dynamic_data_file" in self.config) & 
                    (self.check_dynamic_data.get() == 1)):
                    items.extend(self.globalData['dynamic_data_header'])

                # remove columns with object_id and time
                items = [i for i in items if i != "object_id"]
                items = [i for i in items if i != "time"]
                
                for i in items:
                    self.select_input_feature.insert('END', option=i)
                    self.select_target_feature.insert('END', option=i)
            else:
                if (("dynamic_data_file" in self.config) & 
                    (self.check_dynamic_data.get() == 1)):
                    items = self.globalData['dynamic_data_header'].copy()

                    # remove columns with object_id and time
                    items = [i for i in items if i != "object_id"]
                    items = [i for i in items if i != "time"]
                    
                    for i in items:
                        self.select_input_feature.insert('END', option=i)
                        self.select_target_feature.insert('END', option=i)

        except:
            None
 
    # Get transformer 
    def transform_dynamic_data_option(self, method: str):
        self.config["scaler_input_dynamic_features"] = [method]
    
    def transform_static_data_option(self, method: str):
        self.config["scaler_input_static_features"] = [method]
        
    def transform_target_data_option(self, method: str):
        self.config["scaler_target_features"] = [method]
        
    def data_split_transform(self):
        # Train test split
        # get training and testing periods
        
        try:
            self.config['train_period'] = pd.to_datetime(
                [self.start_train.get_date(), self.end_train.get_date()],
                format = "%Y-%m-%d %H:%M")

            self.config['valid_period'] = pd.to_datetime(
                [self.start_valid.get_date(), self.end_valid.get_date()],
                format = "%Y-%m-%d %H:%M")
            
            self.config['test_period'] = pd.to_datetime(
                [self.start_test.get_date(), self.end_test.get_date()],
                format = "%Y-%m-%d %H:%M")
        except:
            pass
        
        # get input features
        all_items = self.select_input_feature.get(index='all')
        select_index = self.select_input_feature.curselection()
        input_features = [all_items[i] for i in select_index]
        
        # convert target to static and dynamic features
        if "static_data_file" in self.config:
            self.config['input_static_features'] = []
            self.config['input_dynamic_features'] = []
            
            for feature in input_features: 
                if len(self.globalData['static_data_header']) > 0:
                    if feature in self.globalData['static_data_header']:
                        self.config['input_static_features'].append(feature)
                        
                if len(self.globalData['dynamic_data_header']) > 0:
                    if feature in self.globalData['dynamic_data_header']:
                        self.config['input_dynamic_features'].append(feature)
        else:
            self.config['input_dynamic_features'] = input_features
            try: 
                del self.config['input_static_features']
            except:
                pass
        
        # get target features
        all_items = self.select_target_feature.get(index='all')
        select_index = self.select_target_feature.curselection()
        self.config['target_features'] = [all_items[i] for i in select_index]
            
        # get object id
        all_items = self.object_id.get(index='all')
        select_index = self.object_id.curselection()
        self.config['object_id'] = [all_items[i] for i in select_index]
        
        # Get train and test data
        data = read_train_valid_test_data(self.config)
        
        # Update global data
        self.globalData.update(data)
        
        # Delete data variable to free memory
        del data
         
        # Show message box
        print("Done data split (train, valid, test)")
        
        #--------------------------------------------------------Transform data
        # Get scaler name
        self.globalData["x_scaler_method"], \
            self.globalData["y_scaler_method"] = get_scaler_name(self.config)
        print("done get scaler")
        
        # Scale x_train and x_test
        self.globalData["x_scaler"] = Scaler()
        self.globalData["x_scaler"].fit(x=self.globalData["x_train"], 
                                        method=self.globalData["x_scaler_method"])
        
        self.globalData["x_train_scale"] =\
            self.globalData["x_scaler"].transform(x=self.globalData["x_train"])
        print("done x train scale ")    
            
        self.globalData["x_valid_scale"] =\
            self.globalData["x_scaler"].transform(x=self.globalData["x_valid"]) 
        print("done x valid scale")
            
        self.globalData["x_test_scale"] =\
            self.globalData["x_scaler"].transform(x=self.globalData["x_test"])        
        
        # Scale y_train and y_test
        self.globalData["y_scaler"] = Scaler()
        self.globalData["y_scaler"].fit(x=self.globalData["y_train"], 
                                        method=self.globalData["y_scaler_method"])
        self.globalData["y_train_scale"] =\
            self.globalData["y_scaler"].transform(x=self.globalData["y_train"])
        print("done y scale")
        self.globalData["y_valid_scale"] =\
            self.globalData["y_scaler"].transform(x=self.globalData["y_valid"])   
        print("done y valid scale")
        # Show message box
        tk.messagebox.showinfo(title="Message box", 
                               message="Done data split + transform")
        
    def get_first_tensor(self): 
        
        x_train=self.globalData["x_train"].copy()
        x_train_scale=self.globalData["x_train_scale"].copy()
        
        
        if "input_static_features" in self.config: 
            col_names = self.config["input_dynamic_features"] +\
                self.config["input_static_features"]
        else:
            col_names = self.config["input_dynamic_features"]        
        
        x_train = pd.DataFrame(x_train[next(iter(x_train))])
        x_train.columns = [name + "_origignal" for name in col_names]

        x_train_scale = pd.DataFrame(x_train_scale[next(iter(x_train_scale))])
        x_train_scale.columns = [name + "_transform" for name in col_names]
              
        x_train_scale_cat = pd.concat([x_train, x_train_scale], axis=1)
        
        return x_train_scale_cat
        
    def display_orig_trans_data(self):
        try:
            self.table = Table(tk.Toplevel(self), dataframe=self.get_first_tensor(), 
                               showtoolbar=True, showstatusbar=True)
            self.table.show()
        except:
            pass        



from setuptools import setup
from pathlib import Path

# Read the description
readme_file = Path(__file__).absolute().parent / "README.md"
with readme_file.open("r") as fp:
    long_description = fp.read()
    
setup(
    name='hydroecolstm',
    version='0.2.0',    
    description='A python package for HydroEcological Modelling using LSTM',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/tamnva/hydroecolstm',
    author='Tam V. Nguyen',
    author_email='tamnva@gmail.com',
    packages=['hydroecolstm', 
              'hydroecolstm.data',
              'hydroecolstm.interface', 
              'hydroecolstm.utility', 
              'hydroecolstm.model',
              'hydroecolstm.train'],
    python_requires='>=3.8',
    install_requires=['pandas>=2.2.0',
                      'numpy>=1.24.4',
                      'torch>=2.1.0',
                      'pandastable>=0.13.1',
                      'tkcalendar>=1.6.1',
                      'PyYAML>=6.0.1',
                      'pathlib>=1.0.1',
                      'customtkinter>=5.2.1',
                      'CTkToolTip>=0.8',
                      'CTkMessagebox>=2.5',
                      'CTkListbox>=0.10'
                      ],

    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',  
        'Operating System :: POSIX :: Linux',        
        'Programming Language :: Python :: 3',
    ],

    include_package_data=True,
    package_data={'hydroecolstm': ['images/*']},
)




import torch
from torch import nn

import numpy as np



class GaussianMixture(nn.Module):
    def __init__(self, input_size, hidden_size, num_gaussians):
        super(GaussianMixture, self).__init__()
        
        self.z_h = nn.Sequential(
            nn.Linear(input_size, hidden_size),
            nn.Tanh()
        )
        
        self.z_pi = nn.Linear(hidden_size, num_gaussians)
        self.z_sigma = nn.Linear(hidden_size, num_gaussians)
        self.z_mu = nn.Linear(hidden_size, num_gaussians)  

    def forward(self, x):
        z_h = self.z_h(x)
        pi = nn.functional.softmax(self.z_pi(z_h), -1)
        sigma = torch.exp(self.z_sigma(z_h))
        mu = self.z_mu(z_h)
        return pi, sigma, mu

def gaussian_distribution(y, mu, sigma):
    # make |mu|=K copies of y, subtract mu, divide by sigma
    result = (y.expand_as(mu) - mu) * torch.reciprocal(sigma)
    result = -0.5 * (result * result)
    result = (torch.exp(result) * torch.reciprocal(sigma)) * (1.0/np.sqrt(2.0*np.pi))
    return result

def gaussion_loss(pi, sigma, mu, y):
    result = gaussian_distribution(y, mu, sigma) * pi
    result = torch.sum(result, dim=1)
    result = -torch.log(result)
    return torch.mean(result)


# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt # creating visualizations
import numpy as np # basic math and random numbers
import torch # package for building functions with learnable parameters
import torch.nn as nn # prebuilt functions specific to neural networks
from torch.autograd import Variable # storing data while learning
def generate_data(n_samples):
    epsilon = np.random.normal(size=(n_samples))
    x_data = np.random.uniform(-50, 0, n_samples)
    y_data = 7*np.sin(0.75*x_data) + 0.5*x_data + epsilon
    return x_data, y_data
    
n_samples = 1000
x_data, y_data = generate_data(n_samples)

plt.figure(figsize=(8, 8))
plt.scatter(x_data, y_data, alpha=0.2)
plt.show()

n_input = 1
n_hidden = 20
n_output = 1
x_tensor = torch.from_numpy(np.float32(x_data).reshape(n_samples, n_input))
y_tensor = torch.from_numpy(np.float32(y_data).reshape(n_samples, n_input))
x_variable = Variable(x_tensor)
y_variable = Variable(y_tensor, requires_grad=False)
x_test_data = np.linspace(-50, 0, n_samples)

# change data shape, move from numpy to torch
x_test_tensor = torch.from_numpy(np.float32(x_test_data).reshape(n_samples, n_input))
x_test_variable = Variable(x_test_tensor)
x_variable.data = y_tensor
y_variable.data = x_tensor
x_test_data = np.linspace(0, 50, n_samples)
x_test_tensor = torch.from_numpy(np.float32(x_test_data).reshape(n_samples, n_input))
x_test_variable.data = x_test_tensor

mdn_x_data = y_data
mdn_y_data = x_data
mdn_x_tensor = y_tensor
mdn_y_tensor = x_tensor
x_variable = mdn_x_tensor
y_variable = mdn_y_tensor

network = GaussianMixture(1, 20,5)
optimizer = torch.optim.Adam(network.parameters())


for epoch in range(10000):
    pi_variable, sigma_variable, mu_variable = network(x_variable)
    loss = gaussion_loss(pi_variable, sigma_variable, mu_variable, y_variable)
    optimizer.zero_grad()
    loss.backward()
    optimizer.step()
        
    if epoch % 500 == 0:
        print(epoch, loss)

pi_variable, sigma_variable, mu_variable = network(x_test_variable)
pi_data = pi_variable.data.numpy()
sigma_data = sigma_variable.data.numpy()
mu_data = mu_variable.data.numpy()

def gumbel_sample(x, axis=1):
    z = np.random.gumbel(loc=0, scale=1, size=x.shape)
    return (np.log(x) + z).argmax(axis=axis)

k = gumbel_sample(pi_data)

indices = (np.arange(n_samples), k)
rn = np.random.randn(n_samples)
sampled = rn * sigma_data[indices] + mu_data[indices]

plt.figure(figsize=(8, 8))
plt.scatter(mdn_x_data, mdn_y_data, alpha=0.2)
plt.scatter(x_test_data, sampled, alpha=0.2, color='red')
plt.show()



